/* Project MULTIVOL -- input/output from/to multiple volumes with support of
 *                     external scripts
 * main.c -- Argument parsing.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Creation: 08/02/97
 * Update:   08/02/97
 * V2.0 PV001
 * DESCRIPTION
 *    This module is responsible for parsing the command line and storing
 *    configuration information for retrieval by multivol.c
 * NOTES
 * BUGS
 *    - Choosing which direction is subject to EOF is difficult.
 *    - Missing --ignore-errors
 * TODO
 * COPYING-POLICY
 *    GPL
 * BASED-ON
 *    No previous work. First version (V2.0 is the project version).
 * MODIFICATION-HISTORY
 * $Id: main.c,v 1.3 1998/10/21 10:28:00 schaefer Exp $
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include "multivol.h"
#include "multivol_if.h"

/* Private function definitions */

/* NAME
 *    usage
 * DESCRIPTION
 *    Shows this program's usage.
 * RESULT
 *   NONE
 * NOTES
 * BUGS
 * TODO
 */
static void usage(char *prog_name);

/* Function implementation */

/* NAME
 *    main
 * DESCRIPTION
 *    This function parses arguments and executes the
 *    processing functions. We use getopt to parse the
 *    arguments
 * RESULT
 *    0 if ok
 *    1 if minor error
 *    2 if fatal error
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    main.c for explain_inpath V1.0 PV001 25/12/96
 */
int main(int argc, char **argv) {
   multivol_environment_t multivol_env;
   multivol_context_t multivol_context;
   int c;
   BOOL had_fatal_error = FALSE;
   int result = 2; /* FATAL ERROR */

   /* Set up default settings */

   multivol_env.input_file = NULL; /* stdin */
   multivol_env.output_file = NULL; /* stdout */
   multivol_env.tape_init_script = NULL; /* none */
   multivol_env.tape_deinit_script = NULL; /* none */
   multivol_env.medium_changer_script = NULL; /* none */
   multivol_env.block_size = MULTIVOL_DEFAULT_BLOCK_SIZE;
   multivol_env.number_of_tapes = 0; /* infinite */
   multivol_env.to_tape = FALSE; /* for now */
   multivol_env.ignore_error = FALSE;

   multivol_context.current_tape = 0;

   optind = 0; /* global from getopt() */

   while (!had_fatal_error) {
      int option_index = 0;
      static struct option the_long_options[]
         = { { "input-file", TRUE,  FALSE, 'I' },
             { "output-file", TRUE,  FALSE, 'O' },
             { "block-size", TRUE,  FALSE, 'B' },
             { "number-of-tapes", TRUE,  FALSE, 'N' },
             { "tape-init-script", TRUE,  FALSE, 'T' },
             { "tape-deinit-script", TRUE,  FALSE, 'D' },
             { "medium-changer-script", TRUE,  FALSE, 'M' },
#define IGNORE_ERRORS_OPTION 7
             { "ignore-errors", FALSE,  FALSE, 0 },
             { NULL, FALSE, FALSE, 0 } /* END */
           };

      c = getopt_long(argc, argv, "",
                      the_long_options,
                      &option_index);

      if (c == -1) {
         break;
      }

      switch (c) {
         case 0:
#ifdef DEBUG
            fprintf(stderr, "option %s", the_long_options[option_index].name);
            if (optarg) {
               fprintf(stderr, " with arg %s", optarg);
            }
            fprintf(stderr, " ");
#endif
            if ((option_index == IGNORE_ERRORS_OPTION)
                && (optarg == NULL)) {
               multivol_env.ignore_error = TRUE;
            }
            else {
              had_fatal_error = TRUE;
            }
            break;
         case 'I':
#ifdef DEBUG
            fprintf(stderr, "option I with value `%s'", optarg);
#endif
            if (optarg) {
               multivol_env.input_file = strdup(optarg);
               if (multivol_env.input_file == NULL) {
                  fprintf(stderr, "%s: malloc error.\n", argv[0]);
                  had_fatal_error = TRUE;
               }
            }
            else {
               had_fatal_error = TRUE;
            }
            break;
         case 'O':
#ifdef DEBUG
            fprintf(stderr, "option O with value `%s'", optarg);
#endif
            if (optarg) {
               multivol_env.output_file = strdup(optarg);
               if (multivol_env.output_file == NULL) {
                  fprintf(stderr, "%s: malloc error.\n", argv[0]);
                  had_fatal_error = TRUE;
               }
            }
            else {
               had_fatal_error = TRUE;
            }
            break;
         case 'B':
#ifdef DEBUG
            fprintf(stderr, "option B with value `%s'", optarg);
#endif
            if (optarg) {
               multivol_env.block_size = atol(optarg); /* bad */
               if (!(multivol_env.block_size > 0)) {
                  fprintf(stderr, "%s: minimum blocksize is 1.\n",
                                  argv[0]);
                  had_fatal_error = TRUE;
               }
            }
            else {
               had_fatal_error = TRUE;
            }
            break;
         case 'N':
#ifdef DEBUG
            fprintf(stderr, "option N with value `%s'", optarg);
#endif
            if (optarg) {
               multivol_env.number_of_tapes = atol(optarg); /* bad */
            }
            else {
               had_fatal_error = TRUE;
            }
            break;

         case 'T':
#ifdef DEBUG
            fprintf(stderr, "option T with value `%s'", optarg);
#endif
            if (optarg) {
               multivol_env.tape_init_script = strdup(optarg);
               if (multivol_env.tape_init_script == NULL) {
                  fprintf(stderr, "%s: malloc error.\n", argv[0]);
                  had_fatal_error = TRUE;
               }
            }
            else {
               had_fatal_error = TRUE;
            }
            break;
         case 'D':
#ifdef DEBUG
            fprintf(stderr, "option D with value `%s'", optarg);
#endif
            if (optarg) {
               multivol_env.tape_deinit_script = strdup(optarg);
               if (multivol_env.tape_deinit_script == NULL) {
                  fprintf(stderr, "%s: malloc error.\n", argv[0]);
                  had_fatal_error = TRUE;
               }
            }
            else {
               had_fatal_error = TRUE;
            }
            break;
         case 'M':
#ifdef DEBUG
            fprintf(stderr, "option M with value `%s'", optarg);
#endif
            if (optarg) {
               multivol_env.medium_changer_script = strdup(optarg);
               if (multivol_env.medium_changer_script == NULL) {
                  fprintf(stderr, "%s: malloc error.\n", argv[0]);
                  had_fatal_error = TRUE;
               }
            }
            else {
               had_fatal_error = TRUE;
            }
            break;

         case '?':
            had_fatal_error = TRUE;
            break;
         case ':':
            had_fatal_error = TRUE;
            fprintf(stderr, "%s: missing argument.\n", argv[0]);
            break;
         default:
#ifdef DEBUG
            fprintf(stderr, "?? getopt returned character code 0%o ??", c);
#endif
            had_fatal_error = TRUE;
      }
   }

#ifdef DEBUG
   if (optind < argc) {
      fprintf(stderr, "non-option ARGV-elements: ");
   }
   while (optind < argc) {
      fprintf(stderr, "%s ", argv[optind++]);
   }
   fprintf(stderr, " ");
#endif

   if (multivol_env.input_file == NULL) {
      multivol_env.to_tape = TRUE;
   }

   if (had_fatal_error) {
      usage(argv[0]);
      result = 2;
   }
   else {
      result = multivol_do(multivol_env, &multivol_context);
   }

   /* ASSUMPTION
    *    - free(NULL) is ok. See manual page.
    */
   free(multivol_env.input_file);
   free(multivol_env.output_file);
   free(multivol_env.tape_init_script);
   free(multivol_env.tape_deinit_script);
   free(multivol_env.medium_changer_script);

   return result;
}

/* Private function implementation */

static void usage(char *argv) {
   fprintf(stderr, "%s: Usage\n", argv);
   fprintf(stderr, "%s [--input-file=file] [--output-file=file]"
                   "[--block-size=size] [--number-of-tapes=number]"
                   "[--tape-initialization-script=script]"
                   "[--tape-deinitialization-script=script]"
                   "[--medium-changer-script=script] [--ignore-errors]\n",
                   argv);

   fprintf(stderr, "%s [-I file] [-O file] [-B size] [-N number] [-T"
                   " script] [-D script] [-M script]\n",
                   argv);
}
