/* multivol.c -- input/output from/to multiple volumes with support of
 *               external scripts
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Creation: 08/02/97
 * Update:   26/10/98
 * V2.0 PV003
 * DESCRIPTION
 *    This software implements a fixed block output/input with handling
 *    of multiple volumes.
 * USAGE
 *    See manpage
 * BUGS
 *    - stdio/direct mixing maybe.
 *    - result ignored in system().
 * TODO
 *    - Change device name ?
 * ASSUMPTIONS
 *    - We only need to set stdin to raw, and only if input_file == NULL.
 *      They are checked runtime as assertions. Also only if
 *      DONT_SET_TO_RAW is not set. There is infact not much use with
 *      setting to raw.
 * COPYING-POLICY
 *    GPL
 * BASED-ON
 *    My multivol.c V1.0 PV005 of 08/02/97.
 * MODIFICATION-HISTORY
 *    21/10/98  schaefer  Added some comments. Fixed the script execution
 *                        (device and script were inverted, bizarrely).
 *                        Added so that in case of error in write
 *                        we abort read loop.
 *    21/10/98  schaefer  Fixed so that write error (ENOSPC) makes continue
 *                        on next tape without loosing data. Works, but
 *                        data integrity untested yet.
 *    23/10/98  schaefer  Fixed so that we support reading less bytes than
 *                        expected, but still reblocking right. This is needed
 *                        when input to multivol comes from a UNIX pipe.
 *                        This is not implemented in write (but detected if
 *                        it happens ever).
 *    26/10/98  schaefer  Fixed so that short read when infile == tape
 *                        means next tape. Fixed so that EOF on output is
 *                        only when writing to tape.
 * $Id: multivol.c,v 1.7 1998/10/26 13:58:31 schaefer Exp $
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <termio.h>
#include "multivol.h"
#include "multivol_if.h"

/* Private function definitions */

/* NAME
 *    load_next_tape
 * DESCRIPTION
 *    Loads next tape.
 * RESULT
 *    TRUE if everything OK
 * NOTES
 * BUGS
 *    - Before system()ing should set stdin/stdout to stderr.
 *      Else will work only with -I and -O set.
 * TODO
 */
static BOOL load_next_tape(multivol_environment_t multivol_env,
                           multivol_context_t *multivol_context);

#ifndef DONT_SET_TO_RAW
/* NAME
 *    set_raw
 * DESCRIPTION
 *    Sets a specific fd to raw (if mode == TRUE) or cooked mode.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 *    - Can only support upto one channel in raw mode (caches)
 * TODO
 * BASED-ON
 *    My client.c of virtual_connection of 01/03/96.
 */
static void set_raw(BOOL mode, fd_t fd);
#endif

/* Function implementations */

int multivol_do(multivol_environment_t multivol_env,
                multivol_context_t *multivol_context) {
   char *buffer = NULL;

#ifdef DEBUG
   fprintf(stderr, "bs=%ld; total %ld tapes.\n",                   
                   multivol_env.block_size,
                   multivol_env.number_of_tapes);
#endif

   buffer = (char *) malloc(multivol_env.block_size);
   if (buffer) {
      fd_t in_f;
      fd_t out_f;
      BOOL has_failed = FALSE;
      BOOL has_finished = FALSE;
      BOOL partial_write = FALSE;
      multivol_block_size_t partial_buffer_len;

      fflush(stdout);

      while ((((++(multivol_context->current_tape))
               <= multivol_env.number_of_tapes)
              || (multivol_env.number_of_tapes == 0))
             && !has_finished
             && !has_failed) {
         /* NOTES
          *    - First failure ignored.
          */
         has_failed = !load_next_tape(multivol_env, multivol_context);

         if (!partial_write) {
	    if (multivol_env.input_file) {
	       in_f = open(multivol_env.input_file, O_RDONLY);
	    }
	    else {
	       in_f = 0; /* stdin */
#ifndef DONT_SET_TO_RAW
               set_raw(TRUE, in_f);
#endif
            }
         }
	 if (in_f != -1) {
            if (multivol_env.output_file) {
      	       out_f = open(multivol_env.output_file, O_WRONLY);
            }
            else {
               out_f = 1;
            }
            if (out_f != -1) {
	       multivol_block_size_t temp;
	       multivol_block_size_t temp2;

               while (!has_failed && !has_finished) {
                  if (partial_write) {
                     /* NOTES
                      *    - No space on write causes a tape change and
                      *      writing continues.
                      * BUGS
                      *    - Untested yet.
                      */
                     partial_write = FALSE;
                     temp = partial_buffer_len;
                  }
                  else {
                     multivol_block_size_t togo;
                     multivol_block_size_t done;
                     unsigned char *buffer_tmp;

                     /* A short read can happen through a pipe, e.g., where
                      * only PIPE_SIZE bytes (4096) can fit.
                      */

                     togo = multivol_env.block_size;
                     temp = 0;
                     buffer_tmp = buffer;

                     while (togo) {
                        done = read(in_f, buffer_tmp, togo); /* is signed */

			if (done <= 0) {
			   break; /* out of while */
			}
                        else {
                           togo -= done;
                           temp += done;
                           buffer_tmp += done;
                        }
                     }

                     if (temp != multivol_env.block_size) {
                        /* We assume we get the whole block or
                         * nothing (except for last block).
                         */
                        if ((temp == 0) && multivol_env.input_file) {
                            fprintf(stderr, "Assuming EOF on input.\n");
                            break; /* out of while */
                        }
                        else {
                           has_finished = TRUE;
			   fprintf(stderr, "short read %ld\n",
					   temp);
                        }
                     }
                  }

		  if (temp
                      && ((temp2 = write(out_f, buffer, temp)) != temp)) {
		     if (temp2 == -1) {
			perror("Write error");
                        if ((errno == ENOSPC) && multivol_env.output_file) {
                           fprintf(stderr, "Assuming EOT on output.\n");
                           partial_write = TRUE;
                           partial_buffer_len = temp;
                           break; /* out of while */
                        }
                        else {
      			   has_failed = TRUE;
                        }
		     }
		     else {
                        /* Normally, too big write just blocks, and
                         * doesn't just return less bytes.
                         */
			fprintf(stderr, "short write %ld\n",
                                        temp2);
		     }
		  }
	       }

	       if (!has_finished && temp && !partial_write) {
		  perror("Read error");
                  if (!multivol_env.ignore_error) {
   		     has_failed = TRUE;
                  }
	       }
               if (multivol_env.output_file) {
   	          close(out_f);
               }
            }
            else {
	       perror("Can't open output device");
	       has_failed = TRUE;
            }

            if (!partial_write) {
	       if (multivol_env.input_file) {
		  close(in_f);
	       }
	       else {
#ifndef DONT_SET_TO_RAW
                  set_raw(FALSE, in_f);
#endif
               }
            }
            if (multivol_env.to_tape && !partial_write) {
               has_finished = TRUE;
            }
	 }
	 else {
	    perror("Can't open input device");
	    has_failed = TRUE;
	 }
      }

      free(buffer);

      if (has_failed) {
         return 1;
      }
      else {
         return 0;
      }
      /* NOT REACHED */
   }
   else {
      fprintf(stderr, "can't allocate %ld bytes of memory.\n",
              multivol_env.block_size);
      return 1;
   }
   /* NOT REACHED */
}

#define INSERT_NEXT_TAPE_TEXT "\nEnsure next tape loaded in "
static BOOL load_next_tape(multivol_environment_t multivol_env,
                           multivol_context_t *multivol_context) {
   BOOL result = TRUE; /* ok */
   char buffer[MAXBUFLEN];
   char *device_name = multivol_env.to_tape ?
                          multivol_env.output_file
	  		  : multivol_env.input_file;

   /* Case stdin/stdout: EOF is EOF, except for first ``tape'' */
   if (!device_name) {
      return (multivol_context->current_tape == 1);
   }

   if (multivol_env.tape_deinit_script) {      
      /* Call deinitialize script, ignore error */

      if ((strlen(device_name) + 1 + strlen(multivol_env.tape_deinit_script)
          + 1) <= MAXBUFLEN) {
         strcpy(buffer, multivol_env.tape_deinit_script);
         strcat(buffer, " ");
         strcat(buffer, device_name);
#ifdef DEBUG
         fprintf(stderr, "executing %s.\n", buffer);
#endif
         system(buffer); /* result ignored */
      }
      else {
         fprintf(stderr, "device name %s or tape_deinit_script %s too long.\n",
                 device_name, multivol_env.tape_deinit_script);
         result = FALSE;
      }
   }

   /* Either get from user or from medium changer */
   if (result && multivol_env.medium_changer_script) {
      /* Call initialize script */

      if ((strlen(device_name) + 1
	  + strlen(multivol_env.medium_changer_script)
	  + 1
          + 10
          + 1) <= MAXBUFLEN) {
	 strcpy(buffer, multivol_env.medium_changer_script);
	 strcat(buffer, " ");
	 strcat(buffer, device_name);
         sprintf( (buffer + strlen(buffer)),
                  "%10ld", multivol_context->current_tape);
#ifdef DEBUG
         fprintf(stderr, "executing %s.\n", buffer);
#endif
	 result = (system(buffer)) == 0;
      }
      else {
	 fprintf(stderr,
		 "device name %s or medium_changer_script %s too long.\n",
		 device_name, multivol_env.medium_changer_script);
	 result = FALSE;
      }
   }
   else if (result) {
      if (write(2, INSERT_NEXT_TAPE_TEXT, sizeof(INSERT_NEXT_TAPE_TEXT) - 1)
	  == (sizeof(INSERT_NEXT_TAPE_TEXT) - 1)) {
	 if (write(2, device_name, strlen(device_name))
             == strlen(device_name)) {
	    int temp = 0;
	    char c = 0;

	    while (result && (c != 10)) {
	       while ((temp = read(2, &c, 1)) == 0) { ; }
	       if (temp == -1) {
		  fprintf(stderr, "input error.\n");
		  result = FALSE;
	       }
#ifdef DEBUG
	       else {
		  fprintf(stderr, "ignored %d\n", c);
	       }
#endif
	    }

	 }
	 else {
	    fprintf(stderr, "output text error 2.\n");
	    result = FALSE;
	 }
      }
      else {
	 fprintf(stderr, "output text error.\n");
	 result = FALSE;
      }  
   }

   if (result && multivol_env.tape_init_script) {      
      /* Call initialize script */

      if ((strlen(device_name) + 1 + strlen(multivol_env.tape_init_script)
          + 1) <= MAXBUFLEN) {
         strcpy(buffer, multivol_env.tape_init_script);
         strcat(buffer, " ");
         strcat(buffer, device_name);
#ifdef DEBUG
         fprintf(stderr, "executing %s.\n", buffer);
#endif
         result = (system(buffer)) == 0;
      }
      else {
         fprintf(stderr, "device name %s or tape_init_script %s too long.\n",
                 device_name, multivol_env.tape_init_script);
         result = FALSE;
      }
   }
   
   return result;
}

#ifndef DONT_SET_TO_RAW
static struct termio saved;
static fd_t saved_fd = -1;

static void set_raw(BOOL mode, fd_t fd) {
   struct termio tty;

   if (!mode) {
      assert(fd == saved_fd);
      ioctl(fd, TCSETAW, &saved);
      saved_fd = -1;
   }
   else {
      assert(saved_fd == -1);

      ioctl(fd, TCGETA, &tty);
      ioctl(fd, TCGETA, &saved);

      saved_fd = fd;

      tty.c_iflag = (IGNBRK);

      /* No echo, crlf mapping, INTR, QUIT, delays, no erase/kill */
      tty.c_lflag &= ~(ECHO | ICANON | ISIG);

      tty.c_oflag = 0;        /* Transparent output */

      tty.c_cflag &= ~PARENB; /* Same baud rate, disable parity */
      tty.c_cflag |= CS8;     /* Set character size = 8 */
#ifdef MIPS
      tty.c_cflag |= CNEW_RTSCTS; /* MSC94 */
#endif
      tty.c_cc[VMIN] = 1; /* This many chars satisfies reads */
      tty.c_cc[VTIME] = 1;    /* or in this many tenths of seconds */

      ioctl(fd, TCSETAW, &tty);
   }
}
#endif
